require 'uu/os/lang/uniform_hash'

module UU
  module OS
    # Application module
    module Application
      # Class represent application global configuration, configuration is read from config files and environment variables.
      # Configuration is read in this order:
      # * Config file <app-root>/config/app.rb
      # * <uu-home>/config/uu-client.properties
      # * Environment variable SERVER_CFG
      #
      # This mean that you can overwrite default configuration (in ./config/app.rb) by setting same config variable in uu-client.properties.
      #
      # @example
      #   # init config
      #   UU::OS::Application::Config.replace({
      #     :app_credentials => '77-456-2',
      #     :my_param => 'my value'
      #   })
      #   # config params: {:app_credentials => '77-456-2', :my_param => 'my value'}
      #
      #   # get value
      #   UU::OS::Application::Config[:app_credentials] # String '77-456-2'
      #
      #   # merge config
      #   UU::OS::Application::Config.merge!({
      #     :app_credentials => '13-13-2',
      #     :another_param => 13
      #   })
      #   # config params: {:app_credentials => '13-13-2', :my_param => 'my value', :another_param => 13}
      #
      #   # set value
      #   UU::OS::Application::Config[:my_param] # String 'my value'
      #   UU::OS::Application::Config[:my_param] = 'another value'
      #   UU::OS::Application::Config[:my_param] # String 'another value'
      class Config

        # @private
        AUTHN_TOKEN_PARAM = :'uu.os.rest.remote-client-authn-token' # :nodoc:
        private_constant :AUTHN_TOKEN_PARAM

        # @private
        DEFAULT_TTL = 3600 # :nodoc:
        # @private
        @@ctx = UU::OS::Lang::UniformHash.new  # :nodoc:
        # @private
        @@load_callback = nil  # :nodoc:
        # @private
        @@last_update = 0  # :nodoc:
        # @private
        @@ttl = DEFAULT_TTL  # :nodoc:
        # @private
        @@lock = Mutex.new  # :nodoc:
        # @private
        @@reload_disabled = false  # :nodoc:
        # @see Hash#[]
        def self.[](key)
          _reload if reload_necessary?
          @@ctx[key]
        end

        # @see Hash#[]=
        def self.[]=(key, value)
          @@ctx[key] = value
        end

        # @see Hash#merge
        def self.merge!(other_hash, &block)
          @@ctx.merge!(other_hash, &block)
        end

        # @see Hash#replace
        def self.replace(other_hash)
          @@ctx.replace(other_hash)
        end

        # @see Hash#delete
        def self.delete(key, &block)
          @@ctx.delete(key, &block)
        end

        # @see Hash#has_key?
        def self.has_key?(key)
          @@ctx.has_key?(key)
        end

        # @see Hash#clear
        def self.clear
          @@ctx.clear
        end
 
        def self.reload
          _reload(true)
        end

        def self.load(&block)
          return unless block_given?
          @@load_callback = block
        end

        def ttl
          @@ttl
        end

        def ttl=(new_ttl)
          @@ttl = new_ttl
        end

        private

        # @private
        def self._reload(force = false)
          return unless @@load_callback
          return if @@reload_disabled
          @@lock.synchronize do
            return if (!force) && (!reload_necessary?)
            actual_session = Thread.current[AUTHN_TOKEN_PARAM]
            begin
              @@reload_disabled = true
              temp_config = @@ctx.clone
              @@load_callback.call(temp_config)
              @@ctx = temp_config
            rescue Exception => ex
              msg = "Error while executing configuration block. This block is defined by"\
              " UU::OS::Application::Config.load method (typically in <app-root>/config/app.rb file)."
              log_error(ex, msg)
            ensure
              @@reload_disabled = false
              @@last_update = Time.now.to_i
              Thread.current[AUTHN_TOKEN_PARAM] = actual_session
            end
          end
        end

        # @private
        def self.reload_necessary?
          return (@@ttl > 0) && (!@@load_callback.nil?) && ((@@last_update + @@ttl) < Time.now.to_i)
        end

        # @private
        def self.log_error(exception = nil, msg = nil)
          message, exception = exception, message if message.kind_of?(String)
          if (UU::OS.const_defined?(:Logger))
            log_error_uu_log(exception, msg)
          else
            log_error_std_err(exception, msg)
          end
        end

        # @private
        def self.log_error_std_err(exception = nil, message = nil)
          msg = "ERROR #{Time.now} "
          if !message.nil? && !message.empty?
            msg << message.to_s
          end
          msg << "\n  " if !exception.nil?
          if !exception.nil?
            msg << "#{exception.class}"
            msg <<  (exception.respond_to?(:id) ? "[#{exception.id}]:" : " : ")
            msg <<  "#{exception.message}"
            trace = exception.backtrace
            msg << "\n  " << trace.join("\n  ") if !trace.nil? && !trace.empty?
          end
          STDERR.puts(msg)
        end

        # Logs using {UU::OS::Logger}.
        # @private
        def self.log_error_uu_log(exception = nil, message = nil) # :nodoc:
          begin
          @log = UU::OS::Logger.new(self.class)
          @log.error(message, exception)
          rescue Exception => ex
            log_error_std_err(exception, message)
          end
        end

        # instance is not allowed
        private_class_method :new
      end
    end
  end
end